@@ -1,3 +1,4 @@
+{-# LANGUAGE BangPatterns               #-}
 {-# LANGUAGE CPP                        #-}
 {-# LANGUAGE DeriveFunctor              #-}
 {-# LANGUAGE DeriveGeneric              #-}
@@ -5,6 +6,7 @@
 {-# LANGUAGE ExistentialQuantification  #-}
 {-# LANGUAGE FlexibleInstances          #-}
 {-# LANGUAGE GADTSyntax                 #-}
+{-# LANGUAGE LambdaCase                 #-}
 {-# LANGUAGE MultiParamTypeClasses      #-}
 {-# LANGUAGE NamedFieldPuns             #-}
 {-# LANGUAGE NumericUnderscores         #-}
@@ -22,7 +24,6 @@ module Control.Monad.IOSim.Types
   , traceSTM
   , liftST
   , SimA (..)
-  , StepId
   , STMSim
   , STM (..)
   , runSTM
@@ -34,6 +35,7 @@ module Control.Monad.IOSim.Types
   , setCurrentTime
   , unshareClock
   , ScheduleControl (..)
+  , isDefaultSchedule
   , ScheduleMod (..)
   , ExplorationOptions (..)
   , ExplorationSpec
@@ -45,10 +47,12 @@ module Control.Monad.IOSim.Types
   , EventlogEvent (..)
   , EventlogMarker (..)
   , SimEventType (..)
+  , ppSimEventType
   , SimEvent (..)
   , SimResult (..)
+  , ppSimResult
   , SimTrace
-  , Trace.Trace (SimTrace, SimPORTrace, TraceMainReturn, TraceMainException, TraceDeadlock, TraceRacesFound, TraceLoop)
+  , Trace.Trace (SimTrace, SimPORTrace, TraceMainReturn, TraceMainException, TraceDeadlock, TraceRacesFound, TraceLoop, TraceInternalError)
   , ppTrace
   , ppTrace_
   , ppSimEvent
@@ -81,8 +85,7 @@ import qualified Control.Concurrent.Class.MonadSTM.Strict.TVar as StrictTVar
 import           Control.Monad.Class.MonadAsync hiding (Async)
 import qualified Control.Monad.Class.MonadAsync as MonadAsync
 import           Control.Monad.Class.MonadEventlog
-import           Control.Monad.Class.MonadFork hiding (ThreadId)
-import qualified Control.Monad.Class.MonadFork as MonadFork
+import           Control.Monad.Class.MonadFork
 import           Control.Monad.Class.MonadST
 import           Control.Monad.Class.MonadSTM.Internal (MonadInspectSTM (..),
                      MonadLabelledSTM (..), MonadSTM, MonadTraceSTM (..),
@@ -132,6 +135,7 @@ import           Control.Monad.IOSimPOR.Types
 
 
 import qualified System.IO.Error as IO.Error (userError)
+import Data.List (intercalate)
 
 {-# ANN module "HLint: ignore Use readTVarIO" #-}
 newtype IOSim s a = IOSim { unIOSim :: forall r. (a -> SimA s r) -> SimA s r }
@@ -144,7 +148,7 @@ runIOSim (IOSim k) = k Return
 -- `selectTraceEventsDynamic'`.
 --
 traceM :: Typeable a => a -> IOSim s ()
-traceM x = IOSim $ oneShot $ \k -> Output (toDyn x) (k ())
+traceM !x = IOSim $ oneShot $ \k -> Output (toDyn x) (k ())
 
 -- | Trace a value, in the same was as `traceM` does, but from the `STM` monad.
 -- This is primarily useful for debugging.
@@ -158,7 +162,7 @@ data SimA s a where
   Return       :: a -> SimA s a
 
   Say          :: String -> SimA s b -> SimA s b
-  Output       :: Dynamic -> SimA s b -> SimA s b
+  Output       :: !Dynamic -> SimA s b -> SimA s b
 
   LiftST       :: StrictST.ST s a -> (a -> SimA s b) -> SimA s b
 
@@ -180,13 +184,13 @@ data SimA s a where
                   SimA s a -> (e -> SimA s a) -> (a -> SimA s b) -> SimA s b
   Evaluate     :: a -> (a -> SimA s b) -> SimA s b
 
-  Fork         :: IOSim s () -> (ThreadId -> SimA s b) -> SimA s b
-  GetThreadId  :: (ThreadId -> SimA s b) -> SimA s b
-  LabelThread  :: ThreadId -> String -> SimA s b -> SimA s b
+  Fork         :: IOSim s () -> (IOSimThreadId -> SimA s b) -> SimA s b
+  GetThreadId  :: (IOSimThreadId -> SimA s b) -> SimA s b
+  LabelThread  :: IOSimThreadId -> String -> SimA s b -> SimA s b
 
   Atomically   :: STM  s a -> (a -> SimA s b) -> SimA s b
 
-  ThrowTo      :: SomeException -> ThreadId -> SimA s a -> SimA s a
+  ThrowTo      :: SomeException -> IOSimThreadId -> SimA s a -> SimA s a
   SetMaskState :: MaskingState  -> IOSim s a -> (a -> SimA s b) -> SimA s b
   GetMaskState :: (MaskingState -> SimA s b) -> SimA s b
 
@@ -447,7 +451,7 @@ block                a = IOSim (SetMaskState MaskedInterruptible a)
 blockUninterruptible a = IOSim (SetMaskState MaskedUninterruptible a)
 
 instance MonadThread (IOSim s) where
-  type ThreadId (IOSim s) = ThreadId
+  type ThreadId (IOSim s) = IOSimThreadId
   myThreadId       = IOSim $ oneShot $ \k -> GetThreadId k
   labelThread t l  = IOSim $ oneShot $ \k -> LabelThread t l (k ())
 
@@ -455,6 +459,8 @@ instance MonadFork (IOSim s) where
   forkIO task        = IOSim $ oneShot $ \k -> Fork task k
   forkOn _ task      = IOSim $ oneShot $ \k -> Fork task k
   forkIOWithUnmask f = forkIO (f unblock)
+  forkFinally task k = mask $ \restore ->
+                       forkIO $ try (restore task) >>= k
   throwTo tid e      = IOSim $ oneShot $ \k -> ThrowTo (toException e) tid (k ())
   yield              = IOSim $ oneShot $ \k -> YieldSim (k ())
 
@@ -467,6 +473,10 @@ instance MonadSay (STMSim s) where
 
 instance MonadLabelledSTM (IOSim s) where
   labelTVar tvar label = STM $ \k -> LabelTVar label tvar (k ())
+  labelTVarIO tvar label = IOSim $ oneShot $ \k ->
+                                   LiftST ( lazyToStrictST $
+                                            writeSTRef (tvarLabel tvar) $! Just label
+                                          ) k
   labelTQueue  = labelTQueueDefault
   labelTBQueue = labelTBQueueDefault
 
@@ -551,6 +561,10 @@ instance MonadInspectSTM (IOSim s) where
 --
 instance MonadTraceSTM (IOSim s) where
   traceTVar _ tvar f = STM $ \k -> TraceTVar tvar f (k ())
+  traceTVarIO tvar f = IOSim $ oneShot $ \k ->
+                               LiftST ( lazyToStrictST $
+                                        writeSTRef (tvarTrace tvar) $! Just f
+                                      ) k
   traceTQueue  = traceTQueueDefault
   traceTBQueue = traceTBQueueDefault
 
@@ -567,7 +581,15 @@ instance MonadMVar (IOSim s) where
   tryReadMVar  = tryReadMVarDefault
   isEmptyMVar  = isEmptyMVarDefault
 
-data Async s a = Async !ThreadId (STM s (Either SomeException a))
+instance MonadInspectMVar (IOSim s) where
+  type InspectMVarMonad (IOSim s) = ST s
+  inspectMVar p (MVar tvar) = do
+      st <- inspectTVar p tvar
+      case st of
+        MVarEmpty _ _ -> pure Nothing
+        MVarFull x _  -> pure (Just x)
+
+data Async s a = Async !IOSimThreadId (STM s (Either SomeException a))
 
 instance Eq (Async s a) where
     Async tid _ == Async tid' _ = tid == tid'
@@ -734,14 +756,14 @@ data SimEvent
     -- | Used when using `IOSim`.
   = SimEvent {
       seTime        :: !Time,
-      seThreadId    :: !ThreadId,
+      seThreadId    :: !IOSimThreadId,
       seThreadLabel :: !(Maybe ThreadLabel),
       seType        :: !SimEventType
     }
     -- | Only used for /IOSimPOR/
   | SimPOREvent {
       seTime        :: !Time,
-      seThreadId    :: !ThreadId,
+      seThreadId    :: !IOSimThreadId,
       seStep        :: !Int,
       seThreadLabel :: !(Maybe ThreadLabel),
       seType        :: !SimEventType
@@ -759,46 +781,92 @@ ppSimEvent :: Int -- ^ width of the time
            -> Int -- ^ width of thread label
            -> SimEvent
            -> String
-ppSimEvent timeWidth tidWidth tLabelWidth SimEvent {seTime, seThreadId, seThreadLabel, seType} =
+
+ppSimEvent timeWidth tidWidth tLabelWidth SimEvent {seTime = Time time, seThreadId, seThreadLabel, seType} =
     printf "%-*s - %-*s %-*s - %s"
            timeWidth
-           (show seTime)
+           (show time)
            tidWidth
-           (show seThreadId)
+           (ppIOSimThreadId seThreadId)
            tLabelWidth
            threadLabel
-           (show seType)
+           (ppSimEventType seType)
   where
     threadLabel = fromMaybe "" seThreadLabel
-ppSimEvent timeWidth tidWidth tLableWidth SimPOREvent {seTime, seThreadId, seStep, seThreadLabel, seType} =
+
+ppSimEvent timeWidth tidWidth tLableWidth SimPOREvent {seTime = Time time, seThreadId, seStep, seThreadLabel, seType} =
     printf "%-*s - %-*s %-*s - %s"
            timeWidth
-           (show seTime)
+           (show time)
            tidWidth
-           (show (seThreadId, seStep))
+           (ppStepId (seThreadId, seStep))
            tLableWidth
            threadLabel
-           (show seType)
+           (ppSimEventType seType)
   where
     threadLabel = fromMaybe "" seThreadLabel
+
 ppSimEvent _ _ _ (SimRacesFound controls) =
     "RacesFound "++show controls
 
+
 -- | A result type of a simulation.
 data SimResult a
-    = MainReturn    !Time a             ![Labelled ThreadId]
+    = MainReturn    !Time !(Labelled IOSimThreadId) a ![Labelled IOSimThreadId]
     -- ^ Return value of the main thread.
-    | MainException !Time SomeException ![Labelled ThreadId]
+    | MainException !Time !(Labelled IOSimThreadId) SomeException ![Labelled IOSimThreadId]
     -- ^ Exception thrown by the main thread.
-    | Deadlock      !Time               ![Labelled ThreadId]
+    | Deadlock      !Time               ![Labelled IOSimThreadId]
     -- ^ Deadlock discovered in the simulation.  Deadlocks are discovered if
     -- simply the simulation cannot do any progress in a given time slot and
     -- there's no event which would advance the time.
     | Loop
     -- ^ Only returned by /IOSimPOR/ when a step execution took longer than
     -- 'explorationStepTimelimit` was exceeded.
+    | InternalError String
     deriving (Show, Functor)
 
+ppSimResult :: Show a
+            => Int
+            -> Int
+            -> Int
+            -> SimResult a
+            -> String
+ppSimResult timeWidth tidWidth thLabelWidth r = case r of
+    MainReturn (Time time) tid a tids ->
+      printf "%-*s - %-*s %-*s - %s %s"
+             timeWidth
+             (show time)
+             tidWidth
+             (ppIOSimThreadId (l_labelled tid))
+             thLabelWidth
+             (fromMaybe "" $ l_label tid)
+             ("MainReturn " ++ show a)
+             ("[" ++ intercalate "," (ppLabelled ppIOSimThreadId `map` tids) ++ "]")
+    MainException (Time time) tid e tids ->
+      printf "%-*s - %-*s %-*s - %s %s"
+             timeWidth
+             (show time)
+             tidWidth
+             (ppIOSimThreadId (l_labelled tid))
+             thLabelWidth
+             (fromMaybe "" $ l_label tid)
+             ("MainException " ++ show e)
+             ("[" ++ intercalate "," (ppLabelled ppIOSimThreadId `map` tids) ++ "]")
+    Deadlock (Time time) tids ->
+      printf "%-*s - %-*s %-*s - %s %s"
+             timeWidth
+             (show time)
+             tidWidth
+             ""
+             thLabelWidth
+             "" 
+             "Deadlock"
+             ("[" ++ intercalate "," (ppLabelled ppIOSimThreadId `map` tids) ++ "]")
+    Loop -> "<<io-sim-por: step execution exceded explorationStepTimelimit>>"
+    InternalError e -> "<<io-sim internal error: " ++ show e ++ ">>"
+
+
 -- | A type alias for 'IOSim' simulation trace.  It comes with useful pattern
 -- synonyms.
 --
@@ -808,21 +876,21 @@ type SimTrace a = Trace.Trace (SimResult a) SimEvent
 --
 ppTrace :: Show a => SimTrace a -> String
 ppTrace tr = Trace.ppTrace
-               show
-               (ppSimEvent timeWidth tidWith labelWidth)
+               (ppSimResult timeWidth tidWidth labelWidth)
+               (ppSimEvent timeWidth tidWidth labelWidth)
                tr
   where
-    (Max timeWidth, Max tidWith, Max labelWidth) =
+    (Max timeWidth, Max tidWidth, Max labelWidth) =
         bimaximum
       . bimap (const (Max 0, Max 0, Max 0))
               (\a -> case a of
-                SimEvent {seTime, seThreadId, seThreadLabel} ->
-                  ( Max (length (show seTime))
+                SimEvent {seTime = Time time, seThreadId, seThreadLabel} ->
+                  ( Max (length (show time))
                   , Max (length (show (seThreadId)))
                   , Max (length seThreadLabel)
                   )
-                SimPOREvent {seTime, seThreadId, seThreadLabel} ->
-                  ( Max (length (show seTime))
+                SimPOREvent {seTime = Time time, seThreadId, seThreadLabel} ->
+                  ( Max (length (show time))
                   , Max (length (show (seThreadId)))
                   , Max (length seThreadLabel)
                   )
@@ -837,10 +905,10 @@ ppTrace tr = Trace.ppTrace
 ppTrace_ :: SimTrace a -> String
 ppTrace_ tr = Trace.ppTrace
                 (const "")
-                (ppSimEvent timeWidth tidWith labelWidth)
+                (ppSimEvent timeWidth tidWidth labelWidth)
                 tr
   where
-    (Max timeWidth, Max tidWith, Max labelWidth) =
+    (Max timeWidth, Max tidWidth, Max labelWidth) =
         bimaximum
       . bimap (const (Max 0, Max 0, Max 0))
               (\a -> case a of
@@ -859,6 +927,8 @@ ppTrace_ tr = Trace.ppTrace
               )
       $ tr
 
+
+
 -- | Trace each event using 'Debug.trace'; this is useful when a trace ends with
 -- a pure error, e.g. an assertion.
 --
@@ -868,13 +938,13 @@ ppDebug = appEndo
         . Trace.toList
 
 
-pattern SimTrace :: Time -> ThreadId -> Maybe ThreadLabel -> SimEventType -> SimTrace a
+pattern SimTrace :: Time -> IOSimThreadId -> Maybe ThreadLabel -> SimEventType -> SimTrace a
                  -> SimTrace a
 pattern SimTrace time threadId threadLabel traceEvent trace =
     Trace.Cons (SimEvent time threadId threadLabel traceEvent)
                trace
 
-pattern SimPORTrace :: Time -> ThreadId -> Int -> Maybe ThreadLabel -> SimEventType -> SimTrace a
+pattern SimPORTrace :: Time -> IOSimThreadId -> Int -> Maybe ThreadLabel -> SimEventType -> SimTrace a
                     -> SimTrace a
 pattern SimPORTrace time threadId step threadLabel traceEvent trace =
     Trace.Cons (SimPOREvent time threadId step threadLabel traceEvent)
@@ -886,22 +956,25 @@ pattern TraceRacesFound controls trace =
     Trace.Cons (SimRacesFound controls)
                trace
 
-pattern TraceMainReturn :: Time -> a -> [Labelled ThreadId]
+pattern TraceMainReturn :: Time -> Labelled IOSimThreadId -> a -> [Labelled IOSimThreadId]
                         -> SimTrace a
-pattern TraceMainReturn time a threads = Trace.Nil (MainReturn time a threads)
+pattern TraceMainReturn time tid a threads = Trace.Nil (MainReturn time tid a threads)
 
-pattern TraceMainException :: Time -> SomeException -> [Labelled ThreadId]
+pattern TraceMainException :: Time -> Labelled IOSimThreadId -> SomeException -> [Labelled IOSimThreadId]
                            -> SimTrace a
-pattern TraceMainException time err threads = Trace.Nil (MainException time err threads)
+pattern TraceMainException time tid err threads = Trace.Nil (MainException time tid err threads)
 
-pattern TraceDeadlock :: Time -> [Labelled ThreadId]
+pattern TraceDeadlock :: Time -> [Labelled IOSimThreadId]
                       -> SimTrace a
 pattern TraceDeadlock time threads = Trace.Nil (Deadlock time threads)
 
 pattern TraceLoop :: SimTrace a
 pattern TraceLoop = Trace.Nil Loop
 
-{-# COMPLETE SimTrace, SimPORTrace, TraceMainReturn, TraceMainException, TraceDeadlock, TraceLoop #-}
+pattern TraceInternalError :: String -> SimTrace a
+pattern TraceInternalError msg = Trace.Nil (InternalError msg)
+
+{-# COMPLETE SimTrace, SimPORTrace, TraceMainReturn, TraceMainException, TraceDeadlock, TraceLoop, TraceInternalError #-}
 
 
 -- | Events recorded by the simulation.
@@ -916,17 +989,17 @@ data SimEventType
 
   | EventThrow          SomeException
   -- ^ throw exception
-  | EventThrowTo        SomeException ThreadId
+  | EventThrowTo        SomeException IOSimThreadId
   -- ^ throw asynchronous exception (`throwTo`)
   | EventThrowToBlocked
   -- ^ the thread which executed `throwTo` is blocked
   | EventThrowToWakeup
   -- ^ the thread which executed `throwTo` is woken up
-  | EventThrowToUnmasked (Labelled ThreadId)
+  | EventThrowToUnmasked (Labelled IOSimThreadId)
   -- ^ a target thread of `throwTo` unmasked its exceptions, this is paired
   -- with `EventThrowToWakeup` for threads which were blocked on `throwTo`
 
-  | EventThreadForked    ThreadId
+  | EventThreadForked    IOSimThreadId
   -- ^ forked a thread
   | EventThreadFinished
   -- ^ thread terminated normally
@@ -950,7 +1023,7 @@ data SimEventType
                        (Maybe Effect)    -- ^ effect performed (only for `IOSimPOR`)
   | EventTxWakeup      [Labelled TVarId] -- ^ changed vars causing retry
 
-  | EventUnblocked     [ThreadId]
+  | EventUnblocked     [IOSimThreadId]
   -- ^ unblocked threads by a committed STM transaction
 
   --
@@ -962,7 +1035,7 @@ data SimEventType
   | EventThreadDelayFired   TimeoutId
   -- ^ thread woken up after a delay
 
-  | EventTimeoutCreated        TimeoutId ThreadId Time
+  | EventTimeoutCreated        TimeoutId IOSimThreadId Time
   -- ^ new timeout created (via `timeout`)
   | EventTimeoutFired          TimeoutId
   -- ^ timeout fired
@@ -986,8 +1059,8 @@ data SimEventType
   --
 
   -- | event traced when `threadStatus` is executed
-  | EventThreadStatus  ThreadId -- ^ current thread
-                       ThreadId -- ^ queried thread
+  | EventThreadStatus  IOSimThreadId -- ^ current thread
+                       IOSimThreadId -- ^ queried thread
 
   --
   -- /IOSimPOR/ events
@@ -1012,12 +1085,104 @@ data SimEventType
   | EventPerformAction StepId
   -- ^ /IOSimPOR/ event: perform action of the given step
   | EventReschedule           ScheduleControl
+
+  | EventEffect VectorClock Effect
+  -- ^ /IOSimPOR/ event: executed effect; Useful for debugging IOSimPOR or
+  -- showing compact information about thread execution.
+  | EventRaces Races
+  -- ^ /IOSimPOR/ event: races.  Races are updated while we execute
+  -- a simulation.  Useful for debugging IOSimPOR.
   deriving Show
 
+ppSimEventType :: SimEventType -> String
+ppSimEventType = \case
+  EventSay a -> "Say " ++ a
+  EventLog a -> "Dynamic " ++ show a
+  EventMask a -> "Mask " ++ show a
+  EventThrow a -> "Throw " ++ show a
+  EventThrowTo err tid ->
+    concat [ "ThrowTo (",
+              show err, ") ",
+              ppIOSimThreadId tid ]
+  EventThrowToBlocked -> "ThrowToBlocked"
+  EventThrowToWakeup -> "ThrowToWakeup"
+  EventThrowToUnmasked a ->
+    "ThrowToUnmasked " ++ ppLabelled ppIOSimThreadId a
+  EventThreadForked a ->
+    "ThreadForked " ++ ppIOSimThreadId a
+  EventThreadFinished -> "ThreadFinished"
+  EventThreadUnhandled a ->
+    "ThreadUnhandled " ++ show a
+  EventTxCommitted written created mbEff ->
+    concat [ "TxCommitted ",
+             ppList (ppLabelled show) written, " ",
+             ppList (ppLabelled show) created,
+             maybe "" ((' ' :) . ppEffect) mbEff ]
+
+  EventTxAborted mbEff ->
+    concat [ "TxAborted",
+             maybe "" ((' ' :) . ppEffect) mbEff ]
+  EventTxBlocked blocked mbEff ->
+   concat [ "TxBlocked ",
+             ppList (ppLabelled show) blocked,
+             maybe "" ((' ' :) . ppEffect) mbEff ]
+  EventTxWakeup changed ->
+    "TxWakeup " ++ ppList (ppLabelled show) changed
+  EventUnblocked unblocked ->
+    "Unblocked " ++ ppList ppIOSimThreadId unblocked
+  EventThreadDelay tid t ->
+    concat [ "ThreadDelay ",
+             show tid, " ",
+             show t ]
+  EventThreadDelayFired  tid -> "ThreadDelayFired " ++ show tid
+  EventTimeoutCreated timer tid t ->
+    concat [ "TimeoutCreated ",
+             show timer, " ",
+             ppIOSimThreadId tid, " ",
+             show t ]
+  EventTimeoutFired timer ->
+    "TimeoutFired " ++ show timer
+  EventRegisterDelayCreated timer tvarId t ->
+    concat [ "RegisterDelayCreated ",
+             show timer, " ",
+             show tvarId, " ",
+             show t ]
+  EventRegisterDelayFired timer -> "RegisterDelayFired " ++ show timer
+  EventTimerCreated timer tvarId t ->
+    concat [ "TimerCreated ",
+              show timer, " ",
+              show tvarId, " ",
+              show t ]
+  EventTimerUpdated timer t ->
+    concat [ "TimerUpdated ",
+             show timer, " ",
+             show t ]
+  EventTimerCancelled timer -> "TimerCancelled " ++ show timer
+  EventTimerFired timer -> "TimerFired " ++ show timer
+  EventThreadStatus  tid tid' ->
+    concat [ "ThreadStatus ",
+             ppIOSimThreadId tid, " ",
+             ppIOSimThreadId tid' ]
+  EventSimStart a -> "SimStart " ++ show a
+  EventThreadSleep -> "ThreadSleep"
+  EventThreadWake -> "ThreadWake"
+  EventDeschedule a -> "Deschedule " ++ show a
+  EventFollowControl a -> "FollowControl " ++ show a
+  EventAwaitControl s a ->
+    concat [ "AwaitControl ",
+             ppStepId s, " ",
+             show a ]
+  EventPerformAction a -> "PerformAction " ++ ppStepId a
+  EventReschedule a -> "Reschedule " ++ show a
+  EventEffect clock eff ->
+    concat [ "Effect ",
+             ppVectorClock clock, " ",
+             ppEffect eff ]
+  EventRaces a -> show a
 
 -- | A labelled value.
 --
--- For example 'labelThread' or `labelTVar' will insert a label to `ThreadId`
+-- For example 'labelThread' or `labelTVar' will insert a label to `IOSimThreadId`
 -- (or `TVarId`).
 data Labelled a = Labelled {
     l_labelled :: !a,
@@ -1026,6 +1191,10 @@ data Labelled a = Labelled {
   deriving (Eq, Ord, Generic)
   deriving Show via Quiet (Labelled a)
 
+ppLabelled :: (a -> String) -> Labelled a -> String
+ppLabelled pp Labelled { l_labelled = a, l_label = Nothing  } = pp a
+ppLabelled pp Labelled { l_labelled = a, l_label = Just lbl } = concat ["Labelled ", pp a, " ", lbl]
+
 --
 -- Executing STM Transactions
 --
@@ -1045,22 +1214,22 @@ data StmTxResult s a =
        --
        -- It also includes the updated TVarId name supply.
        --
-       StmTxCommitted a [SomeTVar s] -- ^ written tvars
-                        [SomeTVar s] -- ^ read tvars
-                        [SomeTVar s] -- ^ created tvars
-                        [Dynamic]
-                        [String]
-                        TVarId -- updated TVarId name supply
+       StmTxCommitted a ![SomeTVar s] -- ^ written tvars
+                        ![SomeTVar s] -- ^ read tvars
+                        ![SomeTVar s] -- ^ created tvars
+                        ![Dynamic]
+                        ![String]
+                        !TVarId -- updated TVarId name supply
 
        -- | A blocked transaction reports the vars that were read so that the
        -- scheduler can block the thread on those vars.
        --
-     | StmTxBlocked  [SomeTVar s]
+     | StmTxBlocked  ![SomeTVar s]
 
        -- | An aborted transaction reports the vars that were read so that the
        -- vector clock can be updated.
        --
-     | StmTxAborted  [SomeTVar s] SomeException
+     | StmTxAborted  ![SomeTVar s] SomeException
 
 
 -- | A branch indicates that an alternative statement is available in the current
@@ -1094,57 +1263,6 @@ data StmStack s b a where
                    -> StmStack s b c
                    -> StmStack s a c
 
-
----
---- Schedules
----
-
--- | Modified execution schedule.
---
-data ScheduleControl = ControlDefault
-                     -- ^ default scheduling mode
-                     | ControlAwait [ScheduleMod]
-                     -- ^ if the current control is 'ControlAwait', the normal
-                     -- scheduling will proceed, until the thread found in the
-                     -- first 'ScheduleMod' reaches the given step.  At this
-                     -- point the thread is put to sleep, until after all the
-                     -- steps are followed.
-                     | ControlFollow [StepId] [ScheduleMod]
-                     -- ^ follow the steps then continue with schedule
-                     -- modifications.  This control is set by 'followControl'
-                     -- when 'controlTargets' returns true.
-  deriving (Eq, Ord, Show)
-
--- | A schedule modification inserted at given execution step.
---
-data ScheduleMod = ScheduleMod{
-    -- | Step at which the 'ScheduleMod' is activated.
-    scheduleModTarget    :: StepId,
-    -- | 'ScheduleControl' at the activation step.  It is needed by
-    -- 'extendScheduleControl' when combining the discovered schedule with the
-    -- initial one.
-    scheduleModControl   :: ScheduleControl,
-    -- | Series of steps which are executed at the target step.  This *includes*
-    -- the target step, not necessarily as the last step.
-    scheduleModInsertion :: [StepId]
-  }
-  deriving (Eq, Ord)
-
--- | Execution step is identified by the thread id and a monotonically
--- increasing number (thread specific).
---
-type StepId = (ThreadId, Int)
-
-instance Show ScheduleMod where
-  showsPrec d (ScheduleMod tgt ctrl insertion) =
-    showParen (d>10) $
-      showString "ScheduleMod " .
-      showsPrec 11 tgt .
-      showString " " .
-      showsPrec 11 ctrl .
-      showString " " .
-      showsPrec 11 insertion
-
 ---
 --- Exploration options
 ---
@@ -1181,10 +1299,21 @@ data ExplorationOptions = ExplorationOptions{
     -- catching infinite loops etc.
     --
     -- The default value is `Nothing`.
-    explorationReplay        :: Maybe ScheduleControl
+    explorationReplay        :: Maybe ScheduleControl,
     -- ^ A schedule to replay.
     --
     -- The default value is `Nothing`.
+    explorationDebugLevel    :: Int
+    -- ^ Log detailed trace to stderr containing information on discovered
+    -- races.  The trace does not contain the result of the simulation, unless
+    -- one will do that explicitly inside the simulation.
+    --
+    -- level 0: don't show any output,
+    -- level 1: show simulation trace with discovered schedules
+    -- level 2: show simulation trace with discovered schedules and races
+    --
+    -- NOTE: discovered schedules & races are not exposed to the user in the
+    -- callback of `exploreSimTrace` or in the output of `controlSimTrace`.
   }
   deriving Show
 
@@ -1193,7 +1322,8 @@ stdExplorationOptions = ExplorationOptions{
     explorationScheduleBound = 100,
     explorationBranching     = 3,
     explorationStepTimelimit = Nothing,
-    explorationReplay        = Nothing
+    explorationReplay        = Nothing,
+    explorationDebugLevel    = 0
     }
 
 type ExplorationSpec = ExplorationOptions -> ExplorationOptions
